@@ -1,7 +1,16 @@
+/***********************************************************************\
+*   Copyright (C) 1992-1998 by Michael K. Johnson, johnsonm@redhat.com *
+*                                                                      *
+*      This file is placed under the conditions of the GNU Library     *
+*      General Public License, version 2, or any later version.        *
+*      See file ../COPYING for information on distribution conditions. *
+\***********************************************************************/
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <stdlib.h>
 #include <pwd.h>
+#include "proc/procps.h"
 #include <grp.h>
 
 #define	HASHSIZE	16			/* power of 2 */
@@ -25,7 +34,7 @@ char *user_from_uid(uid_t uid)
 	    return((*p)->name);
 	p = &(*p)->next;
     }
-    *p = (struct pwbuf *) malloc(sizeof(struct pwbuf));
+    *p = (struct pwbuf *) xmalloc(sizeof(struct pwbuf));
     (*p)->uid = uid;
     if ((pw = getpwuid(uid)) == NULL)
 	sprintf((*p)->name, "#%d", uid);
@@ -48,16 +57,16 @@ char *group_from_gid(gid_t gid)
 
     g = &grphash[HASH(gid)];
     while (*g) {
-	if ((*g)->gid == gid)
-	    return((*g)->name);
-	g = &(*g)->next;
+       if ((*g)->gid == gid)
+           return((*g)->name);
+       g = &(*g)->next;
     }
     *g = (struct grpbuf *) malloc(sizeof(struct grpbuf));
     (*g)->gid = gid;
     if ((gr = getgrgid(gid)) == NULL)
-	sprintf((*g)->name, "#%d", gid);
+       sprintf((*g)->name, "#%d", gid);
     else
-	sprintf((*g)->name, "%-.8s", gr->gr_name);
+       sprintf((*g)->name, "%-.8s", gr->gr_name);
     (*g)->next = NULL;
     return((*g)->name);
 }
