@@ -1,11 +1,14 @@
 /*
  * New Interface to Process Table -- PROCTAB Stream (a la Directory streams)
- * Copyright(C) 1996. Charles L. Blake.
+ * Copyright (C) 1996 Charles L. Blake.
+ * Copyright (C) 1998 Michael K. Johnson
+ * May be distributed under the conditions of the
+ * GNU Library General Public License; a copy is in COPYING
  */
 #include "proc/version.h"
 #include "proc/readproc.h"
 #include "proc/devname.h"
-#include "proc/ps.h"
+#include "proc/procps.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
@@ -22,26 +25,24 @@
  */
 PROCTAB* openproc(int flags, ...) {
     va_list ap;
-    PROCTAB* PT = malloc(sizeof(PROCTAB));
+    PROCTAB* PT = xmalloc(sizeof(PROCTAB));
 
     if (!Do(PID) && !(PT->procfs = opendir("/proc")))
     	return NULL;
-    if (Do(REAL)) flags |= PROC_FILLSTATUS;
     PT->flags = flags;
     va_start(ap, flags);		/*  Init args list */
-    if (Do(PID)) {
-        PT->pids = va_arg(ap, pid_t*);
-    } else if (Do(NOTPID)) {
-        PT->pids = va_arg(ap, pid_t*);
-        PT->nuid = va_arg(ap, int);
-    } else if (Do(TTY))
+    if (Do(PID))
+    	PT->pids = va_arg(ap, pid_t*);
+    else if (Do(TTY))
     	PT->ttys = va_arg(ap, dev_t*);
-    else if (Do(UID) || Do(NOTUID) || Do(GID) || Do(NOTGID)) {
+    else if (Do(UID)) {
     	PT->uids = va_arg(ap, uid_t*);
 	PT->nuid = va_arg(ap, int);
     } else if (Do(STAT))
     	PT->stats = va_arg(ap, char*);
     va_end(ap);				/*  Clean up args list */
+    if (Do(ANYTTY) && Do(TTY))
+	PT->flags = PT->flags & ~PROC_TTY; /* turn off TTY flag */
     return PT;
 }
 
@@ -66,50 +67,114 @@ void freeproc(proc_t* p) {
     free(p);
 }
 
+
+
 void status2proc (char* S, proc_t* P, int fill) {
     char* tmp;
-
     if (fill == 1) {
         memset(P->cmd, 0, sizeof P->cmd);
-        sscanf (S, "Name:\t%39c", P->cmd);
+        sscanf (S, "Name:\t%15c", P->cmd);
         tmp = strchr(P->cmd,'\n');
         *tmp='\0';
-	tmp = strstr (S,"State");
+        tmp = strstr (S,"State");
         sscanf (tmp, "State:\t%c", &P->state);
     }
+
     tmp = strstr (S,"Pid:");
-    sscanf (tmp, 
-        "Pid:\t%d\nPPid:\t%d\nUid:\t%d\t%d\t%d\t%d\nGid:\t%d\t%d\t%d\t%d\n"
-        "VmSize: %lu kB\nVmLck: %lu kB\nVmRSS: %lu kB\nVmData: %lu kB\n"
-        "VmStk: %lu kB\nVmExe: %lu kB\nVmLib: %lu kB\n"
-        "SigPnd: %lx\nSigBlk: %lx\nSigIgn: %lx\nSigCgt: %lx",
-        &P->pid, &P->ppid, &P->ruid, &P->uid, &P->suid, &P->fsuid,
-        &P->rgid, &P->gid, &P->sgid, &P->fsgid, &P->vm_size, &P->vm_lock,
-	&P->vm_rss, &P->vm_data, &P->vm_stack, &P->vm_exe, &P->vm_lib,
-	&P->signal, &P->blocked, &P->sigignore, &P->sigcatch);
+    if(tmp) sscanf (tmp,
+        "Pid:\t%d\n"
+        "PPid:\t%d\n",
+        &P->pid,
+        &P->ppid
+    );
+    else fprintf(stderr, "Internal error!\n");
+
+    tmp = strstr (S,"Uid:");
+    if(tmp) sscanf (tmp,
+        "Uid:\t%d\t%d\t%d\t%d",
+        &P->ruid, &P->euid, &P->suid, &P->fuid
+    );
+    else fprintf(stderr, "Internal error!\n");
+
+    tmp = strstr (S,"Gid:");
+    if(tmp) sscanf (tmp,
+        "Gid:\t%d\t%d\t%d\t%d",
+        &P->rgid, &P->egid, &P->sgid, &P->fgid
+    );
+    else fprintf(stderr, "Internal error!\n");
+
+    tmp = strstr (S,"VmSize:");
+    if(tmp) sscanf (tmp,
+        "VmSize: %lu kB\n"
+        "VmLck: %lu kB\n"
+        "VmRSS: %lu kB\n"
+        "VmData: %lu kB\n"
+        "VmStk: %lu kB\n"
+        "VmExe: %lu kB\n"
+        "VmLib: %lu kB\n",
+        &P->vm_size, &P->vm_lock, &P->vm_rss, &P->vm_data,
+        &P->vm_stack, &P->vm_exe, &P->vm_lib
+    );
+    else /* looks like an annoying kernel thread */
+    {
+        P->vm_size  = 0;
+        P->vm_lock  = 0;
+        P->vm_rss   = 0;
+        P->vm_data  = 0;
+        P->vm_stack = 0;
+        P->vm_exe   = 0;
+        P->vm_lib   = 0;
+    }
+
+    tmp = strstr (S,"SigPnd:");
+    if(tmp) sscanf (tmp,
+#ifdef SIGNAL_STRING
+        "SigPnd: %s SigBlk: %s SigIgn: %s %*s %s",
+        P->signal, P->blocked, P->sigignore, P->sigcatch
+#else
+        "SigPnd: %Lx SigBlk: %Lx SigIgn: %Lx %*s %Lx",
+        &P->signal, &P->blocked, &P->sigignore, &P->sigcatch
+#endif
+    );
+    else fprintf(stderr, "Internal error!\n");
 }
 
+
+
 /* stat2proc() makes sure it can handle arbitrary executable file basenames
    for `cmd', i.e. those with embedded whitespace or embedded ')'s.  Such names
    confuse %s (see scanf(3)), so the string is split and %39c is used instead.
    (except for embedded ')' "(%[^)]c)" would work.
 */
 void stat2proc(char* S, proc_t* P) {
+    int num;
     char* tmp = strrchr(S, ')');	/* split into "PID (cmd" and "<rest>" */
     *tmp = '\0';			/* replace trailing ')' with NUL */
     /* parse these two strings separately, skipping the leading "(". */
     memset(P->cmd, 0, sizeof P->cmd);	/* clear even though *P xcalloc'd ?! */
-    sscanf(S, "%d (%39c", &P->pid, P->cmd);
-    sscanf(tmp + 2,			/* skip space after ')' too */
-	   "%c %d %d %d %d %d %lu %lu %lu %lu %lu %ld %ld %ld %ld %ld "
-	   "%ld %lu %lu %ld %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu %lu",
-	   &P->state, &P->ppid, &P->pgrp, &P->session, &P->tty, &P->tpgid,
-	   &P->flags, &P->min_flt, &P->cmin_flt, &P->maj_flt, &P->cmaj_flt,
-	   &P->utime, &P->stime, &P->cutime, &P->cstime, &P->priority, &P->nice,
-	   &P->timeout, &P->it_real_value, &P->start_time, &P->vsize, &P->rss,
-	   &P->rss_rlim, &P->start_code, &P->end_code, &P->start_stack,
-	   &P->kstk_esp, &P->kstk_eip, &P->signal, &P->blocked, &P->sigignore,
-	   &P->sigcatch, &P->wchan);
+    sscanf(S, "%d (%15c", &P->pid, P->cmd);   /* comm[16] in kernel */
+    num = sscanf(tmp + 2,			/* skip space after ')' too */
+       "%c "
+       "%d %d %d %d %d "
+       "%lu %lu %lu %lu %lu %lu %lu "
+       "%ld %ld %ld %ld %ld %ld "
+       "%lu %lu "
+       "%ld "
+       "%lu %lu %lu %lu %lu %lu "
+       "%*s %*s %*s %*s " /* discard, no RT signals & Linux 2.1 used hex */
+       "%lu %lu %lu",
+       &P->state,
+       &P->ppid, &P->pgrp, &P->session, &P->tty, &P->tpgid,
+       &P->flags, &P->min_flt, &P->cmin_flt, &P->maj_flt, &P->cmaj_flt, &P->utime, &P->stime,
+       &P->cutime, &P->cstime, &P->priority, &P->nice, &P->timeout, &P->it_real_value,
+       &P->start_time, &P->vsize,
+       &P->rss,
+       &P->rss_rlim, &P->start_code, &P->end_code, &P->start_stack, &P->kstk_esp, &P->kstk_eip,
+/*     P->signal, P->blocked, P->sigignore, P->sigcatch,   */ /* can't use */
+       &P->wchan, &P->nswap, &P->cnswap /* , &P->exit_signal  */);
+/* TODO: add &P->exit_signal support here, perhaps to identify Linux threads */
+    
+/*    fprintf(stderr, "stat2proc converted %d fields.\n",num); */
     if (P->tty == 0)
 	P->tty = -1;  /* the old notty val, update elsewhere bef. moving to 0 */
     if (linux_version_code < LINUX_VERSION(1,3,39)) {
@@ -121,9 +186,11 @@ void stat2proc(char* S, proc_t* P) {
 }
 
 void statm2proc(char* s, proc_t* P) {
-    sscanf(s, "%d %d %d %d %d %d %d",
+    int num;
+    num = sscanf(s, "%ld %ld %ld %ld %ld %ld %ld",
 	   &P->size, &P->resident, &P->share,
 	   &P->trs, &P->lrs, &P->drs, &P->dt);
+/*    fprintf(stderr, "statm2proc converted %d fields.\n",num); */
 }
 
 void nulls2sep(char* str, int len, char sep) {
@@ -167,7 +234,7 @@ char** file2strvec(char* directory, char* what) {
 	}
 	if (end_of_file && buf[n-1])		/* last read char not null */
 	    buf[n++] = '\0';			/* so append null-terminator */
-	rbuf = realloc(rbuf, tot + n);		/* allocate more memory */
+	rbuf = xrealloc(rbuf, tot + n);		/* allocate more memory */
 	memcpy(rbuf + tot, buf, n);		/* copy buffer into it */
 	tot += n;				/* increment total byte ctr */
 	if (end_of_file)
@@ -185,7 +252,7 @@ char** file2strvec(char* directory, char* what) {
 	    c += sizeof(char*);
     c += sizeof(char*);				/* one extra for NULL term */
 
-    rbuf = realloc(rbuf, tot + c + align);	/* make room for ptrs AT END */
+    rbuf = xrealloc(rbuf, tot + c + align);	/* make room for ptrs AT END */
     endbuf = rbuf + tot;			/* addr just past data buf */
     q = ret = (char**) (endbuf+align);		/* ==> free(*ret) to dealloc */
     *q++ = p = rbuf;				/* point ptrs to the strings */
@@ -231,7 +298,6 @@ char** file2strvec(char* directory, char* what) {
  * Unfortunately, the reverse filtering option in which any PID *except* the
  * ones listed is pursued.
  */
-#define flags (PT->flags)
 proc_t* readproc(PROCTAB* PT, proc_t* rbuf) {
     static struct direct *ent;		/* dirent handle */
     static struct stat sb;		/* stat buffer */
@@ -242,6 +308,10 @@ proc_t* readproc(PROCTAB* PT, proc_t* rbuf) {
     /* loop until a proc matching restrictions is found or no more processes */
     /* I know this could be a while loop -- this way is easier to indent ;-) */
 next_proc:				/* get next PID for consideration */
+
+/*printf("PT->flags is 0x%08x\n", PT->flags);*/
+#define flags (PT->flags)
+
     if (Do(PID)) {
 	if (!*PT->pids)			/* set to next item in pids */
 	    return NULL;
@@ -253,105 +323,165 @@ next_proc:				/* get next PID for consideration */
 	    ;
 	if (!ent || !ent->d_name)
 	    return NULL;
-        if (Do(NOTPID) && XinLN (pid_t, (pid_t) atoi(ent->d_name), 
-                                 PT->pids, PT->nuid))
-            goto next_proc;
 	sprintf(path, "/proc/%s", ent->d_name);
     }
     if (stat(path, &sb) == -1)		/* no such dirent (anymore) */
 	goto next_proc;
-    if (!Do(REAL)){
-        if (Do(UID) && !XinLN(uid_t, sb.st_uid, PT->uids, PT->nuid))
-            goto next_proc;		/* not one of the requested uids */
-        if (Do(NOTUID) && XinLN(uid_t, sb.st_uid, PT->uids, PT->nuid))
-            goto next_proc;		/* not one of the requested uids */
-        if (Do(GID) && !XinLN(uid_t, sb.st_gid, PT->uids, PT->nuid))
-            goto next_proc;		/* not one of the requested uids */
-        if (Do(NOTGID) && XinLN(uid_t, sb.st_gid, PT->uids, PT->nuid))
-            goto next_proc;		/* not one of the requested uids */
-    }
+    if (Do(UID) && !XinLN(uid_t, sb.st_uid, PT->uids, PT->nuid))
+	goto next_proc;			/* not one of the requested uids */
 
-    if (!allocated) {			 /* assign mem for return buf */
-	p = rbuf ? rbuf : xcalloc(p, sizeof *p);/* passed buf or alloced mem */
-	allocated = 1;			 /* remember space is set up */
+    if (!allocated) {				 /* assign mem for return buf */
+	p = rbuf ? rbuf : xcalloc(p, sizeof *p); /* passed buf or alloced mem */
+	allocated = 1;				 /* remember space is set up */
     }
-    p->uid = sb.st_uid;			/* need a way to get real uid */
+    p->euid = sb.st_uid;			/* need a way to get real uid */
 
-    if (Do(FILLSTAT)) {
-        if ((file2str(path, "stat", sbuf, sizeof sbuf)) == -1)
-            goto next_proc;               /* error reading /proc/#/stat */
-        stat2proc(sbuf, p);                       /* parse /proc/#/stat */
-        if (Do(FILLSTATUS)) {
-            if ((file2str(path, "status", sbuf, sizeof sbuf)) == -1)
-                goto next_proc;		/* error reading /proc/#/status */
-            status2proc(sbuf, p, 0);            /* parse /proc/#/status */
-        }
-    } else if (Do(FILLSTATUS)) {
-            if ((file2str(path, "status", sbuf, sizeof sbuf)) == -1)
-                goto next_proc;  
-            status2proc(sbuf, p, 1);
-    }
-
-    if (!matched && Do(ANYTTY) && p->tty == -1)
-	goto next_proc;			/* no controlling terminal */
+    if ((file2str(path, "stat", sbuf, sizeof sbuf)) == -1)
+	goto next_proc;			/* error reading /proc/#/stat */
+    stat2proc(sbuf, p);				/* parse /proc/#/stat */
 
     if (!matched && Do(TTY) && !XinL(dev_t, p->tty, PT->ttys))
 	goto next_proc;			/* not one of the requested ttys */
 
+    if (!matched && Do(ANYTTY) && p->tty == -1)
+	goto next_proc;			/* no controlling terminal */
+
     if (!matched && Do(STAT) && !strchr(PT->stats,p->state))
 	goto next_proc;			/* not one of the requested states */
 
-    if (!matched && Do(NOSLEAD) && p->session == p->pid)
-        goto next_proc;
-    
-    if (!matched && Do(NOGLEAD) && p->pgrp == p->pid)
-        goto next_proc;
-
-    if (!matched && Do(REAL)){
-        if (Do(UID)   && !XinLN(uid_t, p->ruid, PT->uids, PT->nuid))
-            goto next_proc;		/* not one of the requested uids */
-        if (Do(NOTUID) && XinLN(uid_t, p->ruid, PT->uids, PT->nuid))
-            goto next_proc;		/* not one of the requested uids */
-        if (Do(GID)   && !XinLN(uid_t, p->rgid, PT->uids, PT->nuid))
-            goto next_proc;		/* not one of the requested uids */
-        if (Do(NOTGID) && XinLN(uid_t, p->rgid, PT->uids, PT->nuid))
-            goto next_proc;		/* not one of the requested uids */
-    }
-    
     if (Do(FILLMEM)) {				/* read, parse /proc/#/statm */
 	if ((file2str(path, "statm", sbuf, sizeof sbuf)) != -1 )
 	    statm2proc(sbuf, p);		/* ignore statm errors here */
     }						/* statm fields just zero */
 
+    if (Do(FILLSTATUS)) {         /* read, parse /proc/#/status */
+       if ((file2str(path, "status", sbuf, sizeof sbuf)) != -1 ){
+           status2proc(sbuf, p, 0 /*FIXME*/);
+       }
+    }
+
     /* some number->text resolving which is time consuming */
-    if (Do(FILLTTY))
-	dev_to_tty(p->ttyc, p->tty);
-    if (Do(FILLUSR)) {
-        strncpy(p->user, user_from_uid(p->uid), sizeof p->user);
-        strncpy(p->group, group_from_gid(p->gid), sizeof p->group);
+    if (Do(FILLUSR)){
+	strncpy(p->euser,   user_from_uid(p->euid), sizeof p->euser);
+        strncpy(p->egroup, group_from_gid(p->egid), sizeof p->egroup);
         if(Do(FILLSTATUS)) {
-            strncpy(p->ruser, user_from_uid(p->ruid), sizeof p->ruser);
-            strncpy(p->suser, user_from_uid(p->suid), sizeof p->suser);
-            strncpy(p->fsuser, user_from_uid(p->fsuid), sizeof p->fsuser);
+            strncpy(p->ruser,   user_from_uid(p->ruid), sizeof p->ruser);
             strncpy(p->rgroup, group_from_gid(p->rgid), sizeof p->rgroup);
+            strncpy(p->suser,   user_from_uid(p->suid), sizeof p->suser);
             strncpy(p->sgroup, group_from_gid(p->sgid), sizeof p->sgroup);
-            strncpy(p->fsgroup, group_from_gid(p->fsgid), sizeof p->fsgroup);
+            strncpy(p->fuser,   user_from_uid(p->fuid), sizeof p->fuser);
+            strncpy(p->fgroup, group_from_gid(p->fgid), sizeof p->fgroup);
         }
     }
-    
-    if (Do(FILLCMD)) {			/* read+parse /proc/#/cmdline */
+
+    if (Do(FILLCMD))				/* read+parse /proc/#/cmdline */
 	p->cmdline = file2strvec(path, "cmdline");
-    }
-    if (Do(FILLENV)) {			/* read+parse /proc/#/environ */
+    if (Do(FILLENV))				/* read+parse /proc/#/environ */
 	p->environ = file2strvec(path, "environ");
-    }
+    
     if (p->state == 'Z')			/* fixup cmd for zombies */
-	strncat(p->cmd," <zombie>", sizeof p->cmd);
+	strncat(p->cmd," <defunct>", sizeof p->cmd);
 
     return p;
 }
 #undef flags
 
+/* ps_readproc: return a pointer to a proc_t filled with requested info about the
+ * next process available matching the restriction set.  If no more such
+ * processes are available, return a null pointer (boolean false).  Use the
+ * passed buffer instead of allocating space if it is non-NULL.  */
+
+/* This is optimized so that if a PID list is given, only those files are
+ * searched for in /proc.  If other lists are given in addition to the PID list,
+ * the same logic can follow through as for the no-PID list case.  This is
+ * fairly complex, but it does try to not to do any unnecessary work.
+ * Unfortunately, the reverse filtering option in which any PID *except* the
+ * ones listed is pursued.
+ */
+proc_t* ps_readproc(PROCTAB* PT, proc_t* rbuf) {
+    static struct direct *ent;		/* dirent handle */
+    static struct stat sb;		/* stat buffer */
+    static char path[32], sbuf[512];	/* bufs for stat,statm */
+    int allocated = 0 /* , matched = 0 */ ;	/* flags */
+    proc_t *p = NULL;
+
+    /* loop until a proc matching restrictions is found or no more processes */
+    /* I know this could be a while loop -- this way is easier to indent ;-) */
+next_proc:				/* get next PID for consideration */
+
+/*printf("PT->flags is 0x%08x\n", PT->flags);*/
+#define flags (PT->flags)
+
+	while ((ent = readdir(PT->procfs)) &&
+	       (*ent->d_name < '0' || *ent->d_name > '9'))
+	    ;
+	if (!ent || !ent->d_name)
+	    return NULL;
+	sprintf(path, "/proc/%s", ent->d_name);
+
+    if (stat(path, &sb) == -1)		/* no such dirent (anymore) */
+	goto next_proc;
+
+    if (!allocated) {				 /* assign mem for return buf */
+	p = rbuf ? rbuf : xcalloc(p, sizeof *p); /* passed buf or alloced mem */
+	allocated = 1;				 /* remember space is set up */
+    }
+    p->euid = sb.st_uid;			/* need a way to get real uid */
+
+    if ((file2str(path, "stat", sbuf, sizeof sbuf)) == -1)
+	goto next_proc;			/* error reading /proc/#/stat */
+    stat2proc(sbuf, p);				/* parse /proc/#/stat */
+
+/*    if (Do(FILLMEM)) {*/				/* read, parse /proc/#/statm */
+	if ((file2str(path, "statm", sbuf, sizeof sbuf)) != -1 )
+	    statm2proc(sbuf, p);		/* ignore statm errors here */
+/*    }			*/			/* statm fields just zero */
+
+  /*  if (Do(FILLSTATUS)) { */        /* read, parse /proc/#/status */
+       if ((file2str(path, "status", sbuf, sizeof sbuf)) != -1 ){
+           status2proc(sbuf, p, 0 /*FIXME*/);
+       }
+/*    }*/
+
+    /* some number->text resolving which is time consuming */
+ /*   if (Do(FILLUSR)){ */
+	strncpy(p->euser,   user_from_uid(p->euid), sizeof p->euser);
+        strncpy(p->egroup, group_from_gid(p->egid), sizeof p->egroup);
+/*        if(Do(FILLSTATUS)) { */
+            strncpy(p->ruser,   user_from_uid(p->ruid), sizeof p->ruser);
+            strncpy(p->rgroup, group_from_gid(p->rgid), sizeof p->rgroup);
+            strncpy(p->suser,   user_from_uid(p->suid), sizeof p->suser);
+            strncpy(p->sgroup, group_from_gid(p->sgid), sizeof p->sgroup);
+            strncpy(p->fuser,   user_from_uid(p->fuid), sizeof p->fuser);
+            strncpy(p->fgroup, group_from_gid(p->fgid), sizeof p->fgroup);
+/*        }*/
+/*    }*/
+
+/*    if (Do(FILLCMD))	*/			/* read+parse /proc/#/cmdline */
+	p->cmdline = file2strvec(path, "cmdline");
+/*    if (Do(FILLENV))	*/			/* read+parse /proc/#/environ */
+	p->environ = file2strvec(path, "environ");
+    
+    if (p->state == 'Z')			/* fixup cmd for zombies */
+	strncat(p->cmd," <defunct>", sizeof p->cmd);
+
+    return p;
+}
+#undef flags
+
+
+void look_up_our_self(proc_t *p) {
+    static char path[32], sbuf[512];	/* bufs for stat,statm */
+    sprintf(path, "/proc/%d", getpid());
+    file2str(path, "stat", sbuf, sizeof sbuf);
+    stat2proc(sbuf, p);				/* parse /proc/#/stat */
+    file2str(path, "statm", sbuf, sizeof sbuf);
+    statm2proc(sbuf, p);		/* ignore statm errors here */
+    file2str(path, "status", sbuf, sizeof sbuf);
+    status2proc(sbuf, p, 0 /*FIXME*/);
+}
+
+
 /* Convenient wrapper around openproc and readproc to slurp in the whole process
  * tree subset satisfying the constraints of flags and the optional PID list.
  * Free allocated memory with freeproctree().  The tree structure is a classic
@@ -385,8 +515,17 @@ proc_t* readproctree(int flags, ...) {
     
     /* pass through apropriate arguments to openproc */
     va_start(ap, flags);
-    if (Do(UID))
-	PT = openproc(flags, va_arg(ap, uid_t*), va_arg(ap, int));
+    if (Do(UID)) {
+	/* temporary variables to ensure that va_arg() instances
+	 * are called in the right order
+	 */
+	uid_t* u;
+	int i;
+
+	u = va_arg(ap, uid_t*);
+	i = va_arg(ap, int);
+	PT = openproc(flags, u, i);
+    }
     else if (Do(PID) || Do(TTY) || Do(STAT))
 	PT = openproc(flags, va_arg(ap, void*));
     else
@@ -411,6 +550,7 @@ proc_t* readproctree(int flags, ...) {
 	    tmp2->r = tmp->l->r;		/* insert as child of found node */
 	    tmp->l->r = node;
 	}
+    closeproc(PT);
     return &tree;
 }
 
@@ -426,17 +566,27 @@ proc_t** readproctab(int flags, ...) {
     va_list ap;
 
     va_start(ap, flags);		/* pass through args to openproc */
-    if (Do(UID))
-	PT = openproc(flags, va_arg(ap, uid_t*), va_arg(ap, int));
+    if (Do(UID)) {
+	/* temporary variables to ensure that va_arg() instances
+	 * are called in the right order
+	 */
+	uid_t* u;
+	int i;
+
+	u = va_arg(ap, uid_t*);
+	i = va_arg(ap, int);
+	PT = openproc(flags, u, i);
+    }
     else if (Do(PID) || Do(TTY) || Do(STAT))
 	PT = openproc(flags, va_arg(ap, void*)); /* assume ptr sizes same */
     else
 	PT = openproc(flags);
     va_end(ap);
     do {					/* read table: */
-	tab = realloc(tab, (n+1)*sizeof(proc_t*));/* realloc as we go, using */
+	tab = xrealloc(tab, (n+1)*sizeof(proc_t*));/* realloc as we go, using */
 	tab[n] = readproc(PT, NULL);		  /* final null to terminate */
     } while (tab[n++]);				  /* stop when NULL reached */
+    closeproc(PT);
     return tab;
 }
 
