@@ -35,9 +35,7 @@
 #include <sys/types.h>
 #include <sys/ioctl.h>
 #include <sys/dir.h>
-#if !defined(__GLIBC__)
-#include <linux/dirent.h>
-#endif /* __GLIBC__ */
+#include <dirent.h>
 
 #define NDEBUG !DEBUG
 
@@ -63,7 +61,7 @@ static char buff[BUFFSIZE]; /* used in the procedures */
 
 int main(int argc, char *argv[]) {
 
-  const char format[]="%2u %1u %1u %5u %5u %5u %5u %3u %3u %4u %4u %4u %4u %3u %3u %3u\n";
+  const char format[]="%2u %2u %2u %6u %6u %6u %6u %3u %3u %5u %5u %4u %5u %3u %3u %3u\n";
   unsigned int height=22; /* window height, reset later if needed. */
   unsigned long int args[2]={0,0};
   unsigned int moreheaders=TRUE;
@@ -79,8 +77,10 @@ int main(int argc, char *argv[]) {
   unsigned int per=0, pero2; 
   unsigned long num=0;
   
-  set_linux_version();
+  setlinebuf(stdout);
   argc=0; /* redefined as number of integer arguments */
+  per=1;
+  num=0;
   for (argv++;*argv;argv++) {
     if ('-' ==(**argv)) {
       switch (*(++(*argv))) {
@@ -95,33 +95,21 @@ int main(int argc, char *argv[]) {
 	/* no other aguments defined yet. */
 	usage();
       }
+    } else {
+      argc++;
+      switch (argc) {
+      case 1:
+        if ((per = atoi(*argv)) == 0)
+         usage();
+       num = ULONG_MAX;
+       break;
+      case 2:
+        num = atol(*argv);
+       break;
+      default:
+       usage();
+      } /* switch */
     }
-    else {
-      if (!sscanf(*argv,"%lu",args+argc++)) usage();
-    }
-  }
-  switch (argc) {
-  case 0:  /* no numeric args */
-    per=1; 
-    num=0;
-    break;
-  case 1:
-    if (args[0])
-	 per=(unsigned) args[0];
-    else
-         per=1;
-    num=ULONG_MAX;
-    break;
-  case 2:
-    if (args[0])
-	 per=(unsigned) args[0];
-    else
-         per=1;
-    num=args[1];
-    break;
-  default : 
-    usage();
-    break;
   }
 
   if (moreheaders) {
@@ -209,9 +197,9 @@ int winhi(void) {
 
 
 void showheader(void){
-  printf("%6s%24s%8s%10s%10s%12s\n",
+  printf("%8s%28s%8s%12s%11s%12s\n",
 	 "procs","memory","swap","io","system","cpu");
-  printf(" %1s %1s %1s %5s %5s %5s %5s %3s %3s %4s %4s %4s %4s %3s %3s %3s\n",
+  printf("%2s %2s %2s %6s %6s %6s %6s %3s %3s %5s %5s %4s %5s %3s %3s %3s\n",
 	 "r","b","w","swpd","free","buff","cache","si","so","bi","bo",
 	 "in","cs","us","sy","id");
 }
