@@ -28,7 +28,7 @@ extern char *ctime ();
 
 char *program_name;
 
-void
+int
 main (argc, argv)
      int argc;
 char *argv[];
@@ -40,11 +40,12 @@ char *argv[];
   int count = 1;
   long timer;
   int nsecs = 2;
+  int total_chars;
 
   program_name = argv[0];
   if (argc < 2)
     {
-      fprintf (stderr, "Usage: %s [-n SLEEP] COMMAND [ARGS ...]\n", argv[0]);
+      fprintf (stderr, "Usage: %s [-n seconds] program [args ...]\n", argv[0]);
       exit (1);
     }
 
@@ -57,19 +58,22 @@ char *argv[];
       count = 3;
       if (nsecs == 0 || argc < 4)
 	{
-	  fprintf (stderr, "Usage: %s [-n SLEEP] COMMAND [ARG ...]\n", argv[0]);
+	  fprintf (stderr, "Usage: %s COMMAND [-n SLEEP] [ARG ...]\n", argv[0]);
 	  exit (1);
 	}
     }
 
   /* Build command string to give to popen.  */
   
+  total_chars = 0;
   bzero (cmd, sizeof (cmd));
-  strcpy (cmd, argv[count]);
-  while (++count < argc)
+  strncpy (cmd, argv[count], sizeof (cmd));
+  total_chars = strlen(argv[count]);
+  while (++count < argc && total_chars < (sizeof(cmd) - 1)) /* 1 for space */
     {
       strcat (cmd, " ");
-      strcat (cmd, argv[count]);
+      strncat (cmd, argv[count], sizeof(cmd) - total_chars);
+      total_chars += strlen(argv[count]) + 1; /* 1 for space */
     }
 
   /* Catch keyboard interrupts so we can put tty back in a sane state.  */
@@ -129,7 +133,7 @@ char *argv[];
   echo ();
   nl ();
   endwin ();
-  exit (0);
+  return 0;
 }
 
 /*------------------------------------.
